import numpy as np
import RobotDART as rd
import dartpy  # OSX breaks if this is imported before RobotDART
import copy
from utils import damped_pseudoinverse, AdT, enforce_joint_limits

class PITask:
    def __init__(self, target, dt, Kp = 10., Ki = 0.1):
        self._target = target
        self._dt = dt
        self._Kp = Kp
        self._Ki = Ki
        self._sum_error = 0
    
    def set_target(self, target):
        self._target = target
    
    # function to compute error
    def error(self, tf):
        return rd.math.logMap(tf.inverse().multiply(self._target))
    
    def update(self, current):
        # error = self._target - current
        Ad_tf = AdT(current)
        error_in_body_frame = self.error(current)
        error_in_world_frame = Ad_tf @ error_in_body_frame

        self._sum_error = self._sum_error + error_in_world_frame * self._dt

        return self._Kp * error_in_world_frame + self._Ki * self._sum_error

# Load a robot
robot = rd.Iiwa()
robot.set_actuator_types("servo")
robot.set_position_enforced(False)

# Load ghost robot for visualization
robot_ghost = robot.clone_ghost()

# set initial joint positions
target_positions = copy.copy(robot.positions())
target_positions[0] = -2.
target_positions[3] = -np.pi / 2.0
target_positions[5] = np.pi / 2.0
robot.set_positions(target_positions)

# get end-effector pose
eef_link_name = "iiwa_link_ee"
tf_desired = robot.body_pose(eef_link_name)

# set robot to random positions
positions = robot.positions()+np.random.rand(robot.num_dofs())*np.pi/1.5-np.pi/3.
robot.set_positions(enforce_joint_limits(robot, positions))

# Create simulator object
dt = 0.001
simu = rd.RobotDARTSimu(dt)

# Create Graphics
gconfig = rd.gui.GraphicsConfiguration(1024, 768) # Create a window of 1024x768 resolution/size
graphics = rd.gui.Graphics(gconfig) # create graphics object with configuration
simu.set_graphics(graphics)
graphics.look_at([3., 1., 2.], [0., 0., 0.])
# graphics.record_video("task-space.mp4")

# Add robot and nice floor
simu.add_robot(robot)
robot_ghost.set_positions(target_positions)
simu.add_robot(robot_ghost)
simu.add_checkerboard_floor()

Kp = 2. # Kp could be an array of 6 values
Ki = 0.01 # Ki could be an array of 6 values
controller = PITask(tf_desired, dt, Kp, Ki)

# Run simulation
while True:
    if simu.step_world():
        break
    tf = robot.body_pose(eef_link_name)
    vel = controller.update(tf)
    jac = robot.jacobian(eef_link_name) # this is in world frame
    jac_pinv = damped_pseudoinverse(jac) # np.linalg.pinv(jac) # get pseudo-inverse
    cmd = jac_pinv @ vel

    # alpha = 2.
    # cmd = alpha * (jac.T @ vel) # using jacobian transpose
    robot.set_commands(cmd)
